from unittest import mock

import swapper
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.core.management import call_command
from django.db import transaction
from django.test import TransactionTestCase
from django.urls import reverse
from django.utils.module_loading import import_string
from openwisp_ipam.tests import CreateModelsMixin as SubnetIpamMixin

from openwisp_controller.config.tests.utils import (
    CreateConfigTemplateMixin,
    TestVpnX509Mixin,
    TestWireguardVpnMixin,
    TestZeroTierVpnMixin,
)
from openwisp_network_topology.tests.utils import CreateGraphObjectsMixin
from openwisp_users.tests.utils import TestOrganizationMixin
from openwisp_utils.admin_theme.dashboard import DASHBOARD_CHARTS, DASHBOARD_TEMPLATES
from openwisp_utils.admin_theme.menu import MENU

from ..base.models import logger as models_logger
from ..base.models import trigger_device_checks_path

Node = swapper.load_model('topology', 'Node')
Link = swapper.load_model('topology', 'Link')
Topology = swapper.load_model('topology', 'Topology')
DeviceNode = swapper.load_model('topology_device', 'DeviceNode')
WifiMesh = swapper.load_model('topology_device', 'WifiMesh')
Device = swapper.load_model('config', 'Device')
Template = swapper.load_model('config', 'Template')
Vpn = swapper.load_model('config', 'Template')
Organization = swapper.load_model('openwisp_users', 'Organization')
Cert = swapper.load_model('pki', 'Cert')


class Base(
    TestVpnX509Mixin,
    TestWireguardVpnMixin,
    TestZeroTierVpnMixin,
    SubnetIpamMixin,
    CreateConfigTemplateMixin,
    CreateGraphObjectsMixin,
    TestOrganizationMixin,
):
    topology_model = Topology
    node_model = Node
    _ZT_SERVICE_REQUESTS = 'openwisp_controller.config.api.zerotier_service.requests'
    _ZT_GENERATE_IDENTITY_SUBPROCESS = 'openwisp_controller.config.base.vpn.subprocess'

    def _init_test_node(
        self,
        topology,
        addresses=None,
        label='test',
        common_name=None,
        create=True,
    ):
        if not addresses:
            addresses = ['netjson_id']
        node = Node(
            organization=topology.organization,
            topology=topology,
            label=label,
            addresses=addresses,
            properties={'common_name': common_name},
        )
        if create:
            node.full_clean()
            node.save()
        return node

    def _init_wireguard_test_node(self, topology, addresses=[], create=True, **kwargs):
        if not addresses:
            addresses = ['public_key']
        properties = {
            'preshared_key': None,
            'endpoint': None,
            'latest_handsake': '0',
            'transfer_rx': '0',
            'transfer_tx': '0',
            'persistent_keepalive': 'off',
            'allowed_ips': ['10.0.0.2/32'],
        }
        properties.update(kwargs)
        allowed_ips = properties.get('allowed_ips')
        node = Node(
            organization=topology.organization,
            topology=topology,
            label=','.join(allowed_ips),
            addresses=addresses,
            properties=properties,
        )
        if create:
            node.full_clean()
            node.save()
        return node

    def _init_zerotier_test_node(
        self, topology, addresses=None, label='test', zt_member_id=None, create=True
    ):
        if not addresses:
            addresses = [self._TEST_ZT_MEMBER_CONFIG['address']]
        node = Node(
            organization=topology.organization,
            topology=topology,
            label=label,
            addresses=addresses,
            # zt peer address is `zt_memeber_id`
            properties={'address': zt_member_id},
        )
        if create:
            node.full_clean()
            node.save()
        return node

    def _create_wireguard_test_env(self, parser):
        org = self._get_org()
        device, _, _ = self._create_wireguard_vpn_template()
        device.organization = org
        topology = self._create_topology(organization=org, parser=parser)
        return topology, device

    @mock.patch(_ZT_GENERATE_IDENTITY_SUBPROCESS)
    @mock.patch(_ZT_SERVICE_REQUESTS)
    def _create_zerotier_test_env(self, mock_requests, mock_subprocess, parser):
        mock_requests.get.side_effect = [
            # For node status
            self._get_mock_response(200, response=self._TEST_ZT_NODE_CONFIG)
        ]
        mock_requests.post.side_effect = [
            # For create network
            self._get_mock_response(200),
            # For controller network join
            self._get_mock_response(200),
            # For controller auth and ip assignment
            self._get_mock_response(200),
            # For member auth and ip assignment
            self._get_mock_response(200),
        ]
        mock_stdout = mock.MagicMock()
        mock_stdout.stdout.decode.return_value = self._TEST_ZT_MEMBER_CONFIG['identity']
        mock_subprocess.run.return_value = mock_stdout
        org = self._get_org()
        device, _, _ = self._create_zerotier_vpn_template()
        device.organization = org
        topology = self._create_topology(organization=org, parser=parser)
        zerotier_member_id = device.config.vpnclient_set.first().zerotier_member_id
        return topology, device, zerotier_member_id

    def _create_test_env(self, parser):
        organization = self._get_org()
        vpn = self._create_vpn(name='test VPN', organization=organization)
        self._create_template(
            name='VPN',
            type='vpn',
            vpn=vpn,
            config=vpn.auto_client(),
            default=True,
            organization=organization,
        )
        vpn2 = self._create_vpn(name='test VPN2', ca=vpn.ca, organization=organization)
        self._create_template(
            name='VPN2',
            type='vpn',
            vpn=vpn2,
            config=vpn.auto_client(),
            default=True,
            organization=organization,
        )
        device = self._create_device(organization=organization)
        config = self._create_config(device=device)
        topology = self._create_topology(organization=organization, parser=parser)
        cert = config.vpnclient_set.first().cert
        return topology, device, cert


class TestControllerIntegration(Base, TransactionTestCase):
    def test_auto_create_openvpn(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        self.assertEqual(DeviceNode.objects.count(), 0)
        with self.subTest('assert number of queries'):
            with self.assertNumQueries(15):
                node = self._init_test_node(topology, common_name=cert.common_name)
        self.assertEqual(DeviceNode.objects.count(), 1)
        device_node = DeviceNode.objects.first()
        self.assertEqual(device_node.device, device)
        self.assertEqual(device_node.node, node)

        with self.subTest('not run on save'):
            with mock.patch.object(transaction, 'on_commit') as on_commit:
                node.save()
                on_commit.assert_not_called()

    def test_auto_create_openvpn_failures(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')

        with self.subTest('common_name not present'):
            self._init_test_node(topology)
            self.assertFalse(DeviceNode.objects.exists())

        with self.subTest('cert does not exist'):
            self._init_test_node(topology, common_name='missing')
            self.assertFalse(DeviceNode.objects.exists())

        with self.subTest('exception during save'):
            with mock.patch.object(
                DeviceNode, 'save', side_effect=Exception('test')
            ) as save:
                with mock.patch.object(models_logger, 'exception') as logger_exception:
                    self._init_test_node(topology, common_name=cert.common_name)
                    save.assert_called_once()
                    logger_exception.assert_called_once()
                    self.assertEqual(DeviceNode.objects.count(), 0)

    def test_auto_create_wireguard(self):
        topology, device = self._create_wireguard_test_env(
            parser='netdiff.WireguardParser'
        )
        self.assertEqual(DeviceNode.objects.count(), 0)
        with self.subTest('return if node has no allowed ips'):
            node = self._init_wireguard_test_node(topology, allowed_ips=[])
            self.assertEqual(DeviceNode.objects.count(), 0)
        with self.subTest('handle error if node has bogus allowed ips'):
            try:
                node = self._init_wireguard_test_node(topology, allowed_ips=['invalid'])
            except ValueError:
                self.fail('ValueError raised')
        with self.subTest('assert number of queries'):
            with self.assertNumQueries(15):
                node = self._init_wireguard_test_node(topology)
        self.assertEqual(DeviceNode.objects.count(), 1)
        device_node = DeviceNode.objects.first()
        self.assertEqual(device_node.device, device)
        self.assertEqual(device_node.node, node)
        with self.subTest('do not raise Exception on link status changed'):
            try:
                target_node = self._init_wireguard_test_node(topology)
                link = Link(
                    topology=node.topology, source=node, target=target_node, cost=0
                )
                link.save()
                link.status = 'down'
                link.save(update_fields=['status'])
            except KeyError:
                self.fail('KeyError raised')

    def test_auto_create_zerotier(self):
        topology, device, zerotier_member_id = self._create_zerotier_test_env(
            parser='netdiff.ZeroTierParser'
        )
        self.assertEqual(DeviceNode.objects.count(), 0)
        with self.subTest('assert number of queries'):
            with self.assertNumQueries(15):
                node = self._init_zerotier_test_node(
                    topology, zt_member_id=zerotier_member_id
                )
        self.assertEqual(DeviceNode.objects.count(), 1)
        device_node = DeviceNode.objects.first()
        self.assertEqual(device_node.device, device)
        self.assertEqual(device_node.node, node)

        with self.subTest('not run on save'):
            with mock.patch.object(transaction, 'on_commit') as on_commit:
                node.save()
                on_commit.assert_not_called()

    def test_auto_create_zerotier_failures(self):
        topology, device, zerotier_member_id = self._create_zerotier_test_env(
            parser='netdiff.ZeroTierParser'
        )

        with self.subTest('zerotier_member_id not present'):
            self._init_zerotier_test_node(topology)
            self.assertFalse(DeviceNode.objects.exists())

        with self.subTest('zerotier_member_id does not exist'):
            self._init_zerotier_test_node(topology, zt_member_id='non_existent_id')
            self.assertFalse(DeviceNode.objects.exists())

        with self.subTest('exception during save'):
            with mock.patch.object(
                DeviceNode, 'save', side_effect=Exception('test')
            ) as save:
                with mock.patch.object(models_logger, 'exception') as logger_exception:
                    self._init_zerotier_test_node(
                        topology, zt_member_id=zerotier_member_id
                    )
                    save.assert_called_once()
                    logger_exception.assert_called_once()
                    self.assertEqual(DeviceNode.objects.count(), 0)

    def test_filter_by_link(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')

        with self.subTest('empty result'):
            node = self._init_test_node(topology, common_name=cert.common_name)
            self.assertFalse(DeviceNode.filter_by_link(Link()).exists())

        with self.subTest('non empty result'):
            node2 = self._init_test_node(
                topology, addresses=['netjson_id2'], label='test2'
            )
            link = Link(
                source=node,
                target=node2,
                topology=topology,
                organization=topology.organization,
                cost=1,
            )
            link.full_clean()
            link.save()
            self.assertTrue(DeviceNode.filter_by_link(link).exists())

    def test_link_down_openvpn(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        node = self._init_test_node(topology, common_name=cert.common_name)
        node2 = self._init_test_node(topology, addresses=['netjson_id2'], label='test2')
        device.management_ip = '10.0.0.8'
        device.save()
        link = Link(
            source=node,
            target=node2,
            status='up',
            topology=topology,
            organization=topology.organization,
            cost=1,
        )
        link.full_clean()
        link.save()
        with self.assertNumQueries(6):
            link.status = 'down'
            link.save()
        device.refresh_from_db()
        self.assertIsNone(device.management_ip)

    def test_link_up_openvpn(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        node = self._init_test_node(
            topology, addresses=['netjson_id', '10.0.0.2'], common_name=cert.common_name
        )
        node2 = self._init_test_node(topology, addresses=['netjson_id2'], label='test2')
        device.management_ip = None
        device.save()
        link = Link(
            source=node,
            target=node2,
            status='down',
            topology=topology,
            organization=topology.organization,
            cost=1,
        )
        link.full_clean()
        link.save()
        with self.assertNumQueries(6):
            link.status = 'up'
            link.save()
        device.refresh_from_db()
        self.assertEqual(device.management_ip, '10.0.0.2')

    @mock.patch.object(models_logger, 'warning')
    def test_link_up_openvpn_failure(self, logger_warning):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        node = self._init_test_node(
            topology,
            addresses=['netjson_id', 'not_an_ip'],
            common_name=cert.common_name,
        )
        node2 = self._init_test_node(topology, addresses=['netjson_id2'], label='test2')
        device.management_ip = None
        device.save()
        link = Link(
            source=node,
            target=node2,
            status='down',
            topology=topology,
            organization=topology.organization,
            cost=1,
        )
        link.full_clean()
        link.save()

        with self.subTest('Test invalid addresses'):
            link.status = 'up'
            link.save()
            logger_warning.assert_called_once()
            logger_warning.assert_called_with(
                'ValueError raised while processing addresses: netjson_id, not_an_ip'
            )

        link.status = 'down'
        link.save()
        logger_warning.reset_mock()

        with self.subTest('Test disconnected nodes'):
            node.addresses = ['10.0.0.1']
            node.save()
            link.status = 'up'
            link.save()
            logger_warning.assert_called_once()
            logger_warning.assert_called_with(
                'IndexError raised while processing addresses: 10.0.0.1'
            )

        device.refresh_from_db()
        self.assertIsNone(device.management_ip)

    def test_node_label_override(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        node = self._init_test_node(topology, common_name=cert.common_name)
        node.refresh_from_db()
        self.assertEqual(node.get_name(), device.name)

    def test_topology_json(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        self._init_test_node(topology, common_name=cert.common_name)
        with self.subTest('assert number of queries'):
            with self.assertNumQueries(2):
                json = topology.json(dict=True)
        self.assertEqual(json['nodes'][0]['label'], device.name)

    def test_create_device_nodes_command(self):
        topology, _, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        properties = {'common_name': cert.common_name}
        n = self._create_node(topology=topology, properties=properties)
        DeviceNode.objects.all().delete()
        call_command('create_device_nodes')
        qs = DeviceNode.objects.filter(node=n)
        self.assertEqual(qs.count(), 1)

    def test_shared_topology_org_devices(self):
        org1 = self._create_org(name='org1')
        org2 = self._create_org(name='org2')
        shared_topology = self._create_topology(
            organization=None, parser='netdiff.OpenvpnParser'
        )
        shared_vpn = self._create_vpn(name='test VPN', organization=None)
        self._create_template(
            name='VPN',
            type='vpn',
            vpn=shared_vpn,
            config=shared_vpn.auto_client(),
            default=True,
            organization=None,
        )
        org1_device = self._create_device(organization=org1)
        self._create_config(device=org1_device)
        org1_cert_common_name = (
            org1_device.config.vpnclient_set.first().cert.common_name
        )
        org2_device = self._create_device(organization=org2)
        self._create_config(device=org2_device)
        org2_cert_common_name = (
            org2_device.config.vpnclient_set.first().cert.common_name
        )

        openvpn_status = (
            'OpenVPN CLIENT LIST\n'
            'Updated,Sun Oct  8 19:46:06 2017\n'
            'Common Name,Real Address,Bytes Received,Bytes Sent,Connected Since\n'
            f'{org1_cert_common_name},87.7.9.213:56857,417209,6046898,Sun Oct  8 08:39:11 2017\n'
            f'{org2_cert_common_name},2.226.154.66:52091,6020304,444463,Sun Oct  8 08:39:10 2017\n'
            'ROUTING TABLE\n'
            'Virtual Address,Common Name,Real Address,Last Ref\n'
            f'9e:c2:f2:55:f4:33,{org1_cert_common_name},87.7.9.213:56857,Sun Oct  8 19:45:37 2017\n'
            f'46:8a:69:e7:46:24,{org2_cert_common_name},2.226.154.66:52091,Sun Oct  8 19:45:37 2017\n'
            'GLOBAL STATS\n'
            'Max bcast/mcast queue length,6\n'
            'END'
        )
        shared_topology.receive(openvpn_status)
        self.assertEqual(DeviceNode.objects.count(), 2)
        org1_node = DeviceNode.objects.get(device=org1_device).node
        org2_node = DeviceNode.objects.get(device=org2_device).node
        self.assertEqual(org1_node.organization, org1)
        self.assertEqual(org2_node.get_organization_id(), org2.id)


class TestMonitoringIntegration(Base, TransactionTestCase):
    @mock.patch(
        'openwisp_monitoring.device.apps.DeviceMonitoringConfig.connect_device_signals'
    )
    @mock.patch('openwisp_monitoring.db.timeseries_db.create_database')
    @mock.patch('openwisp_monitoring.device.utils.manage_short_retention_policy')
    @mock.patch('openwisp_monitoring.device.utils.manage_default_retention_policy')
    @mock.patch('openwisp_monitoring.device.exportable._exportable_fields', [])
    @mock.patch('openwisp_notifications.types.NOTIFICATION_TYPES', dict())
    def test_monitoring_integration(self, *args):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        node = self._init_test_node(
            topology, addresses=['netjson_id', '10.0.0.2'], common_name=cert.common_name
        )
        node2 = self._init_test_node(topology, addresses=['netjson_id2'], label='test2')
        device.management_ip = None
        device.save()
        link = Link(
            source=node,
            target=node2,
            status='up',
            topology=topology,
            organization=topology.organization,
            cost=1,
        )
        link.full_clean()
        link.save()
        DASHBOARD_CHARTS.clear()
        DASHBOARD_TEMPLATES.clear()
        MENU.clear()
        # needed for monitoring
        trigger_device_checks = import_string(trigger_device_checks_path)
        with mock.patch.object(trigger_device_checks, 'delay') as mocked_task:
            with self.modify_settings(
                INSTALLED_APPS={
                    'append': [
                        'openwisp_controller.connection',
                        'openwisp_controller.geo',
                        'openwisp_monitoring.monitoring',
                        'openwisp_monitoring.device',
                        'openwisp_monitoring.check',
                        'openwisp_notifications',
                    ]
                }
            ):
                link.status = 'down'
                link.save()
            mocked_task.assert_called_once()
            mocked_task.assert_called_with(device.pk, recovery=False)


class TestAdmin(Base, TransactionTestCase):
    module = 'openwisp_network_topology'
    app_label = 'topology'
    topology_model = Topology
    link_model = Link
    node_model = Node
    user_model = get_user_model()
    fixtures = ['test_users.json']
    api_urls_path = 'api.urls'

    @property
    def prefix(self):
        return 'admin:{0}'.format(self.app_label)

    def setUp(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        node1 = self._init_test_node(topology, common_name=cert.common_name)
        node2 = self._init_test_node(topology, addresses=['netjson_id2'], label='test2')
        link = Link(
            source=node1,
            target=node2,
            topology=topology,
            organization=topology.organization,
            cost=1,
        )
        link.full_clean()
        link.save()
        self._create_link(topology=topology, source=node1, target=node2)
        self.client.force_login(self.user_model.objects.get(username='admin'))

    def test_node_change_list_queries(self):
        path = reverse('{0}_node_changelist'.format(self.prefix))
        with self.assertNumQueries(5):
            self.client.get(path)

    def test_link_change_list_queries(self):
        path = reverse('{0}_link_changelist'.format(self.prefix))
        with self.assertNumQueries(5):
            self.client.get(path)

    def test_link_node_different_topology(self):
        Topology.objects.all().delete()
        org = self._create_org()
        topology1 = self._create_topology(organization=org)
        topology2 = self._create_topology(organization=org)
        node1 = self._init_test_node(topology2)
        node2 = self._init_test_node(topology2)
        link = Link(
            source=node1,
            target=node2,
            topology=topology1,
            organization=topology1.organization,
            cost=1,
        )
        with self.assertRaises(ValidationError) as context:
            # Raises ValidationError if link belongs to diff topologies
            link.full_clean()
        self.assertIn('source', context.exception.error_dict)
        self.assertIn('target', context.exception.error_dict)

    def test_link_update_status(self):
        t = Topology.objects.first()
        n1 = self._create_node(label='node1org1', topology=t)
        n2 = self._create_node(label='node2org1', topology=t)
        link = self._create_link(topology=t, source=n1, target=n2)
        path = reverse('{0}_link_change'.format(self.prefix), args=[link.pk])
        response = self.client.post(
            path,
            data={
                'topology': str(t.pk),
                'organization': str(t.organization_id),
                'source': str(n1.id),
                'target': str(n2.id),
                'cost': '1.0',
                'status': 'down',
            },
            follow=True,
        )
        self.assertEqual(response.status_code, 200)
        link.refresh_from_db()
        self.assertEqual(link.status, 'down')

    def test_topology_admin(self):
        """
        Tests WifiMeshInlineAdmin is absent in TopologyAdmin
        when OPENWISP_NETWORK_TOPOLOGY_WIFI_MESH_INTEGRATION
        is set to False.

        Note: This test is present here because TopologyAdmin class
        cannot be patched based on app_settings.WIFI_MESH_INTEGRATION
        once the project is initialized.
        """
        topology = Topology.objects.first()
        response = self.client.get(
            reverse(f'{self.prefix}_topology_change', args=[topology.id])
        )
        self.assertNotContains(response, 'Wifi mesh')

    def test_topology_get_name_desc(self):
        response = self.client.get(reverse(f'{self.prefix}_node_changelist'))
        self.assertNotContains(response, '<span>Get name</span>', html=True)
        self.assertContains(response, '<span>Label</span>', html=True)



# Generated by Django 3.1.14 on 2022-02-02 07:12

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('topology', '0013_add_user_defined_properties_field'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='snapshot',
            name='organization',
        ),
    ]



SIMPLE_MESH_DATA = {
    '64:70:02:c3:03:b2': [
        {
            'mac': '64:70:02:c3:03:b3',
            'mtu': 1500,
            'multicast': True,
            'name': 'mesh0',
            'txqueuelen': 1000,
            'type': 'wireless',
            'up': True,
            'wireless': {
                'channel': 11,
                'clients': [
                    {
                        'auth': True,
                        'authorized': True,
                        'ht': True,
                        'mac': 'a4:bc:3f:ae:c7:0c',
                        'mfp': False,
                        'noise': -90,
                        'vendor': 'TP-LINK TECHNOLOGIES CO.,LTD.',
                        'vht': False,
                        'wmm': True,
                        'mesh_llid': 19000,
                        'mesh_plid': 24000,
                        'mesh_non_peer_ps': 'ACTIVE',
                    },
                    {
                        'auth': True,
                        'authorized': True,
                        'ht': True,
                        'mac': '2a:9a:fb:12:11:77',
                        'mfp': False,
                        'noise': -94,
                        'signal': -58,
                        'vendor': 'TP-LINK TECHNOLOGIES CO.,LTD.',
                        'vht': False,
                        'wmm': True,
                        'mesh_llid': 19500,
                        'mesh_plid': 24500,
                        'mesh_non_peer_ps': 'ACTIVE',
                    },
                ],
                'country': 'ES',
                'frequency': 2462,
                'htmode': 'HT20',
                'mode': '802.11s',
                'noise': -95,
                'signal': -61,
                'ssid': 'Test Mesh',
                'tx_power': 17,
            },
        }
    ],
    'a4:bc:3f:ae:c7:0b': [
        {
            'mac': 'a4:bc:3f:ae:c7:0c',
            'mtu': 1500,
            'multicast': True,
            'name': 'mesh0',
            'txqueuelen': 1000,
            'type': 'wireless',
            'up': True,
            'wireless': {
                'channel': 11,
                'clients': [
                    {
                        'auth': True,
                        'authorized': True,
                        'ht': True,
                        'mac': '64:70:02:C3:03:B3',
                        'mfp': False,
                        'noise': -98,
                        'signal': -58,
                        'vendor': 'TP-LINK TECHNOLOGIES CO.,LTD.',
                        'vht': False,
                        'wmm': True,
                        'mesh_llid': 20000,
                        'mesh_plid': 25000,
                        'mesh_non_peer_ps': 'ACTIVE',
                    },
                    {
                        'auth': True,
                        'authorized': True,
                        'ht': True,
                        'mac': '2a:9a:fb:12:11:77',
                        'mfp': False,
                        'noise': -96,
                        'signal': -60,
                        'vendor': 'TP-LINK TECHNOLOGIES CO.,LTD.',
                        'vht': False,
                        'wmm': True,
                        'mesh_llid': 19500,
                        'mesh_plid': 24500,
                        'mesh_non_peer_ps': 'ACTIVE',
                    },
                ],
                'country': 'ES',
                'frequency': 2462,
                'htmode': 'HT20',
                'mode': '802.11s',
                'noise': -95,
                'signal': -56,
                'ssid': 'Test Mesh',
                'tx_power': 17,
            },
        }
    ],
    '2a:9a:fb:12:11:76': [
        {
            'mac': '2a:9a:fb:12:11:77',
            'mtu': 1500,
            'multicast': True,
            'name': 'mesh0',
            'txqueuelen': 1000,
            'type': 'wireless',
            'up': True,
            'wireless': {
                'channel': 11,
                'clients': [
                    {
                        # Link properties
                        'auth': True,
                        'authorized': True,
                        'noise': -92,
                        'signal': -56,
                        # Node properties
                        'ht': True,
                        'mfp': False,
                        'vht': False,
                        'wmm': True,
                        'mac': 'a4:bc:3f:ae:c7:0c',
                        'vendor': 'TP-LINK TECHNOLOGIES CO.,LTD.',
                        'mesh_non_peer_ps': 'LISTEN',
                    },
                    {
                        'mac': '0a:cc:ae:34:ff:3d',
                        'wps': False,
                        'wds': False,
                        'ht': True,
                        'preauth': False,
                        'assoc': True,
                        'authorized': True,
                        'vht': False,
                        'wmm': False,
                        'aid': 1,
                        'mfp': False,
                        'auth': True,
                        'vendor': 'TP-LINK TECHNOLOGIES CO.,LTD.',
                        'signature': 'test_signature',
                        'mesh_plink': 'LISTEN',
                    },
                ],
                'country': 'ES',
                'frequency': 2462,
                'htmode': 'HT20',
                'mode': '802.11s',
                'noise': -95,
                'signal': -64,
                'ssid': 'Test Mesh',
                'tx_power': 17,
            },
        },
        {
            'addresses': [
                {
                    'address': '192.168.1.1',
                    'family': 'ipv4',
                    'mask': 24,
                    'proto': 'static',
                },
            ],
            'bridge_members': ['eth0.1', 'mesh0'],
            'mac': '2a:9a:fb:12:11:76',
            'mtu': 1500,
            'multicast': True,
            'name': 'br-lan',
            'speed': '-1F',
            'stp': False,
            'txqueuelen': 1000,
            'type': 'bridge',
            'up': True,
        },
        {
            'name': 'wlan0',
            'type': 'wireless',
            'up': True,
            'mac': '2a:9a:fb:12:11:78',
            'txqueuelen': 1000,
            'multicast': True,
            'mtu': 1500,
            'wireless': {
                'frequency': 2437,
                'mode': 'access_point',
                'signal': -29,
                'tx_power': 6,
                'channel': 6,
                'ssid': 'testnet',
                'noise': -95,
                'country': 'US',
                'htmode': 'HT20',
                'clients': [
                    {
                        'mac': '00:ee:ad:34:f5:3b',
                        'wps': False,
                        'wds': False,
                        'ht': True,
                        'preauth': False,
                        'assoc': True,
                        'authorized': True,
                        'vht': False,
                        'wmm': True,
                        'aid': 1,
                        'mfp': False,
                        'auth': True,
                        'vendor': 'TP-LINK TECHNOLOGIES CO.,LTD.',
                        'signature': 'test_signature',
                    },
                ],
            },
        },
    ],
}

SINGLE_NODE_MESH_DATA = {
    '64:70:02:c3:03:b2': [
        {
            'mac': '64:70:02:c3:03:b3',
            'mtu': 1500,
            'multicast': True,
            'name': 'mesh0',
            'txqueuelen': 1000,
            'type': 'wireless',
            'up': True,
            'wireless': {
                'channel': 11,
                'clients': [],
                'country': 'ES',
                'frequency': 2462,
                'htmode': 'HT20',
                'mode': '802.11s',
                'noise': -95,
                'signal': -61,
                'ssid': 'Test Mesh',
                'tx_power': 17,
                'mesh_llid': 19751,
                'mesh_local_ps': 'ACTIVE',
                'mesh_non_peer_ps': 'ACTIVE',
                'mesh_peer_ps': 'ACTIVE',
                'mesh_plid': 24413,
                'mesh_plink': 'ESTAB',
            },
        }
    ],
}



import json
import logging

import swapper
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from netdiff.exceptions import NetdiffException
from rest_framework import generics, pagination
from rest_framework.authentication import BasicAuthentication, SessionAuthentication
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from openwisp_users.api.authentication import BearerAuthentication
from openwisp_users.api.mixins import FilterByOrganizationManaged, ProtectedAPIMixin
from openwisp_users.api.permissions import DjangoModelPermissions, IsOrganizationManager

from .. import settings as app_settings
from ..utils import get_object_or_404
from .filters import LinkFilter, NetworkCollectionFilter, NodeFilter
from .parsers import TextParser
from .serializers import (
    LinkSerializer,
    NetworkGraphSerializer,
    NetworkGraphUpdateSerializer,
    NodeSerializer,
)

logger = logging.getLogger(__name__)
Snapshot = swapper.load_model('topology', 'Snapshot')
Topology = swapper.load_model('topology', 'Topology')
Node = swapper.load_model('topology', 'Node')
Link = swapper.load_model('topology', 'Link')


class RequireAuthentication(APIView):
    if app_settings.TOPOLOGY_API_AUTH_REQUIRED:
        authentication_classes = [
            SessionAuthentication,
            BasicAuthentication,
            BearerAuthentication,
        ]
        permission_classes = [
            IsAuthenticated,
            IsOrganizationManager,
            DjangoModelPermissions,
        ]


class ListViewPagination(pagination.PageNumberPagination):
    page_size = 10
    page_size_query_param = 'page_size'
    max_page_size = 100


class UnpublishedTopologyFilterMixin:
    def get_queryset(self):
        qs = super().get_queryset()
        if self.request.query_params.get('include_unpublished'):
            return qs
        return qs.filter(published=True)


class NetworkCollectionView(
    UnpublishedTopologyFilterMixin,
    RequireAuthentication,
    FilterByOrganizationManaged,
    generics.ListCreateAPIView,
):
    """
    Data of all the topologies returned
    in NetJSON NetworkCollection format.
    """

    serializer_class = NetworkGraphSerializer
    queryset = Topology.objects.select_related('organization')
    filter_backends = (DjangoFilterBackend,)
    filterset_class = NetworkCollectionFilter

    def list(self, request, *args, **kwargs):
        self.check_permissions(request)
        return super().list(request, *args, **kwargs)

    def get_success_headers(self, data):
        """
        Remove `LOCATION` option from the
        header when URL field is present.
        """
        return {}


class NetworkGraphView(
    UnpublishedTopologyFilterMixin,
    RequireAuthentication,
    FilterByOrganizationManaged,
    generics.RetrieveUpdateDestroyAPIView,
):
    """
    Data of a specific topology returned
    in NetJSON NetworkGraph format.
    """

    serializer_class = NetworkGraphUpdateSerializer
    queryset = Topology.objects.select_related('organization')


class ReceiveTopologyView(APIView):
    """
    This views allow nodes to send topology data using the RECEIVE strategy.
    Required query string parameters:
        * key
    Allowed content-types:
        * text/plain
    """

    model = Topology
    parser_classes = (TextParser,)

    def _validate_request(self, request, topology):
        key = request.query_params.get('key')
        # wrong content type: 415
        if request.content_type != 'text/plain':
            return Response(
                {'detail': _('expected content type "text/plain"')}, status=415
            )
        # missing key: 400
        if not key:
            return Response(
                {'detail': _('missing required "key" parameter')}, status=400
            )
        # wrong key 403
        if topology.key != key:
            return Response({'detail': _('wrong key')}, status=403)
        return

    def post(self, request, pk, format=None):
        topology = get_object_or_404(self.model, pk, strategy='receive')
        validation_response = self._validate_request(request, topology)
        if validation_response:
            return validation_response
        try:
            topology.receive(request.data)
        except NetdiffException as e:
            error = _(
                'Supplied data not recognized as %s, '
                'got exception of type "%s" '
                'with message "%s"'
            ) % (topology.get_parser_display(), e.__class__.__name__, e)
            return Response({'detail': error}, status=400)
        success_message = _('data received successfully')
        return Response({'detail': success_message})


class NetworkGraphHistoryView(RequireAuthentication):
    """
    History of a specific topology returned
    in NetJSON NetworkGraph format.
    """

    topology_model = Topology
    snapshot_model = Snapshot
    queryset = topology_model.objects.all()  # Required for DjangoModelPermissions

    def get(self, request, pk, format=None):
        topology = get_object_or_404(self.topology_model, pk)
        self.check_object_permissions(request, topology)
        date = request.query_params.get('date')
        options = dict(topology=topology, date=date)
        # missing date: 400
        if not date:
            return Response(
                {'detail': _('missing required "date" parameter')}, status=400
            )
        try:
            s = self.snapshot_model.objects.get(**options)
            return Response(json.loads(s.data))
        except self.snapshot_model.DoesNotExist:
            return Response(
                {'detail': _('no snapshot found for this date')}, status=404
            )
        except ValidationError:
            return Response({'detail': _('invalid date supplied')}, status=403)


class NodeListCreateView(
    ProtectedAPIMixin, FilterByOrganizationManaged, generics.ListCreateAPIView
):
    queryset = Node.objects.order_by('-created')
    serializer_class = NodeSerializer
    pagination_class = ListViewPagination
    filter_backends = (DjangoFilterBackend,)
    filterset_class = NodeFilter


class NodeDetailView(
    ProtectedAPIMixin,
    FilterByOrganizationManaged,
    generics.RetrieveUpdateDestroyAPIView,
):
    queryset = Node.objects.all().select_related('topology')
    serializer_class = NodeSerializer


class LinkListCreateView(
    ProtectedAPIMixin, FilterByOrganizationManaged, generics.ListCreateAPIView
):
    queryset = Link.objects.select_related(
        'topology',
        'source',
        'target',
    ).order_by('-created')
    serializer_class = LinkSerializer
    pagination_class = ListViewPagination
    filter_backends = (DjangoFilterBackend,)
    filterset_class = LinkFilter


class LinkDetailView(
    ProtectedAPIMixin,
    FilterByOrganizationManaged,
    generics.RetrieveUpdateDestroyAPIView,
):
    queryset = Link.objects.select_related(
        'topology',
        'source',
        'target',
    )
    serializer_class = LinkSerializer


network_collection = NetworkCollectionView.as_view()
network_graph = NetworkGraphView.as_view()
network_graph_history = NetworkGraphHistoryView.as_view()
receive_topology = ReceiveTopologyView.as_view()
node_list = NodeListCreateView.as_view()
node_detail = NodeDetailView.as_view()
link_list = LinkListCreateView.as_view()
link_detail = LinkDetailView.as_view()



from swapper import swappable_setting

from .base.models import AbstractDeviceNode, AbstractWifiMesh


class DeviceNode(AbstractDeviceNode):
    class Meta(AbstractDeviceNode.Meta):
        abstract = False
        swappable = swappable_setting('topology_device', 'DeviceNode')


class WifiMesh(AbstractWifiMesh):
    class Meta(AbstractWifiMesh.Meta):
        abstract = False
        swappable = swappable_setting('topology_device', 'WifiMesh')



from django.contrib import admin
from swapper import load_model

from openwisp_network_topology.admin import TopologyAdmin

from . import settings as app_settings

WifiMesh = load_model('topology_device', 'WifiMesh')
Topology = load_model('topology', 'Topology')


class WifiMeshInlineAdmin(admin.StackedInline):
    model = WifiMesh
    extra = 0
    max_num = 1


if app_settings.WIFI_MESH_INTEGRATION:
    TopologyAdmin.inlines = list(TopologyAdmin.inlines) + [WifiMeshInlineAdmin]



from django.urls import include, path

from .api import urls as api
from .visualizer import urls as visualizer_urls

urlpatterns = [
    path('accounts/', include('openwisp_users.accounts.urls')),
    path('api/v1/', include(api)),
    path('topology/', include(visualizer_urls)),
]



import json

from asgiref.sync import async_to_sync
from channels.generic.websocket import WebsocketConsumer
from django.core.exceptions import ValidationError
from swapper import load_model

from . import settings as app_settings

Topology = load_model('topology', 'Topology')


class TopologyConsumer(WebsocketConsumer):
    channel_layer_group = 'topology'

    def _is_user_authorized_to_view_topology(self, user, topology_pk):
        try:
            topology = Topology.objects.get(pk=topology_pk)
        except (Topology.DoesNotExist, ValidationError):
            return False
        if not app_settings.TOPOLOGY_API_AUTH_REQUIRED:
            return True
        return user.is_superuser or (
            user.is_authenticated
            and user.is_manager(topology.organization_id)
            and user.has_perm(f'{Topology._meta.app_label}.view_topology')
        )

    def connect(self):
        user, topology_pk = self.scope.get('user'), self.scope.get('url_route').get(
            'kwargs'
        ).get('pk')
        if self._is_user_authorized_to_view_topology(user, topology_pk):
            async_to_sync(self.channel_layer.group_add)(
                f"{self.channel_layer_group}-{topology_pk}", self.channel_name
            )
            self.accept()
        else:
            self.close()

    def disconnect(self, close_code):
        self.close()
        pass

    def send_topology_update(self, event):
        self.send(
            text_data=json.dumps(
                {
                    'type': "broadcast_topology",
                    'topology': event['data'],
                }
            )
        )



# Generated by Django 4.0.1 on 2022-01-31 18:06

import django.db.models.deletion
from django.db import migrations, models
from swapper import get_model_name


class Migration(migrations.Migration):
    dependencies = [
        ('topology', '0014_remove_snapshot_organization'),
    ]

    operations = [
        migrations.AlterField(
            model_name='link',
            name='organization',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=get_model_name('openwisp_users', 'Organization'),
                verbose_name='organization',
            ),
        ),
        migrations.AlterField(
            model_name='node',
            name='organization',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=get_model_name('openwisp_users', 'Organization'),
                verbose_name='organization',
            ),
        ),
        migrations.AlterField(
            model_name='topology',
            name='organization',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=get_model_name('openwisp_users', 'Organization'),
                verbose_name='organization',
            ),
        ),
    ]



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _
from swapper import get_model_name

from openwisp_utils.admin_theme.menu import register_menu_group

from .settings import SIGNALS


class OpenwispNetworkTopologyConfig(AppConfig):
    name = 'openwisp_network_topology'
    label = 'topology'
    verbose_name = _('Network Topology')

    def ready(self, *args, **kwargs):
        if SIGNALS:  # pragma: nocover
            __import__(SIGNALS)
        from .signals import broadcast_topology, update_topology

        update_topology.connect(broadcast_topology)
        self.register_menu_groups()

    def register_menu_groups(self):
        register_menu_group(
            position=110,
            config={
                'label': 'Network Topology',
                'items': {
                    1: {
                        'label': _('Topologies'),
                        'model': get_model_name('topology', 'Topology'),
                        'name': 'changelist',
                        'icon': 'ow-topology',
                    },
                    2: {
                        'label': _('Nodes'),
                        'model': get_model_name('topology', 'Node'),
                        'name': 'changelist',
                        'icon': 'ow-node',
                    },
                    3: {
                        'label': _('Links'),
                        'model': get_model_name('topology', 'Link'),
                        'name': 'changelist',
                        'icon': 'ow-link',
                    },
                },
                'icon': 'ow-network-topology',
            },
        )



from django.urls import path

from . import consumers

websocket_urlpatterns = [
    # This route is used by both
    # the admin and non-admin topology view
    path(
        'ws/network-topology/topology/<uuid:pk>/',
        consumers.TopologyConsumer.as_asgi(),
    ),
]



import logging

from celery import shared_task
from django.core.exceptions import ObjectDoesNotExist
from swapper import load_model

logger = logging.getLogger(__name__)


@shared_task
def handle_update_topology(topology_pk, diff):
    """
    A Celery task that updates the network topology
    of a Topology instance in the background.

    Args:
        topology_pk (uuid):
        The primary key of the Topology instance.

        diff (str):
        A dict containing the network topology diff.
    """
    Topology = load_model('topology', 'Topology')
    try:
        topology = Topology.objects.get(pk=topology_pk)
    except ObjectDoesNotExist as e:
        logger.warning(f'handle_update_topology("{topology_pk}") failed: {e}')
        return
    topology.update_topology(diff)



#!/usr/bin/env python
from setuptools import find_packages, setup

from openwisp_network_topology import get_version


def get_install_requires():
    """
    parse requirements.txt, ignore links, exclude comments
    """
    requirements = []
    for line in open('requirements.txt').readlines():
        # skip to next iteration if comment or empty line
        if (
            line.startswith('#')
            or line == ''
            or line.startswith('http')
            or line.startswith('git')
        ):
            continue
        # add line to requirements
        requirements.append(line)
    return requirements


setup(
    name='openwisp-network-topology',
    version=get_version(),
    license='BSD-3-Clause',
    author='OpenWISP',
    author_email='support@openwisp.io',
    description='OpenWISP Network Topology',
    long_description=open('README.rst').read(),
    url='http://openwisp.org',
    download_url='https://github.com/openwisp/openwisp-network-topology/releases',
    platforms=['Platform Independent'],
    keywords=['django', 'netjson', 'openwrt', 'networking', 'openwisp'],
    packages=find_packages(exclude=['tests*', 'docs*']),
    include_package_data=True,
    zip_safe=False,
    install_requires=get_install_requires(),
    classifiers=[
        'Development Status :: 5 - Production/Stable ',
        'Environment :: Web Environment',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: System :: Networking',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Framework :: Django',
        'Programming Language :: Python :: 3',
    ],
)



from django.conf import settings
from django.contrib.auth.models import Permission
from django.db import migrations
from swapper import dependency, split

from . import create_default_permissions


def assign_permissions_to_groups(apps, schema_editor):
    create_default_permissions(apps, schema_editor)

    def _add_permission_to_group(group, models, operations):
        for model in models:
            for operation in operations:
                try:
                    permission = Permission.objects.get(
                        codename='{}_{}'.format(operation, model)
                    )
                except Permission.DoesNotExist:
                    continue
                else:
                    group.permissions.add(permission.pk)

    operators_can_manage = ['link', 'node']
    operators_can_only_view = ['topology']
    admin_can_manage = ['link', 'node', 'topology']
    manage_operations = ['add', 'change', 'delete', 'view']
    view_only_operations = ['view']
    Group = apps.get_model('openwisp_users', 'Group')

    try:
        admin = Group.objects.get(name='Administrator')
        operator = Group.objects.get(name='Operator')
    # consider failures custom cases
    # that do not have to be dealt with
    except Group.DoesNotExist:
        return

    # Add permissions for operators
    _add_permission_to_group(operator, operators_can_manage, manage_operations)
    _add_permission_to_group(operator, operators_can_only_view, view_only_operations)
    # Add permissions for Administrator
    _add_permission_to_group(admin, admin_can_manage, manage_operations)


class Migration(migrations.Migration):
    dependencies = [
        dependency(*split(settings.AUTH_USER_MODEL), version='0004_default_groups'),
        ('topology', '0004_fixed_target_link_set'),
    ]
    operations = [
        migrations.RunPython(
            assign_permissions_to_groups, reverse_code=migrations.RunPython.noop
        )
    ]
